// Copyright (c) 2015, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

library omnibus.message_builder;

import 'dart:io';
import 'dart:convert';

import 'package:args/args.dart';
import 'package:yaml/yaml.dart';
import 'package:path/path.dart' as path;
import 'package:dart_style/dart_style.dart';

/// Omnibus message builder
///
/// This executable generates Dart code from YAML Omnibus message
/// specifications. More details can be found in `doc/omnibuild.md`.
main(List<String> args) {
  // Create CLI args parser.
  var parser = new ArgParser();
  parser.addOption('output', abbr: 'o', help: 'Output directory.');

  // Parse CLI args.
  final options = parser.parse(args);
  String outputDir = options['output'];

  // Read package name.
  String packageName =
      loadYaml(new File('./pubspec.yaml').readAsStringSync())['name'];

  // Loop through input files.
  options.rest.forEach((String filepath) {
    // Generate code.
    var code = _generateMessageBuilders(
        new File(filepath).readAsStringSync(), packageName);

    // Write to output file.
    new File(outputDir + path.basenameWithoutExtension(filepath) + '.dart')
        .writeAsStringSync(code);
  });
}

/// Apply transformer to a single file.
String _generateMessageBuilders(String yamlData, String packageName) {
  // Read YAML file.
  YamlMap yaml = loadYaml(yamlData);
  var code = new StringBuffer();

  // Write copyright.
  if (yaml.containsKey('copyright') && yaml['copyright'].isNotEmpty) {
    code.write(_prefixEachLine(yaml['copyright'], '// '));
    code.writeln();
  }

  // Write auto-generated notice.
  code.writeln('// THIS FILE IS AUTO-GENERATED, PLEASE DO NOT MODIFY BY HAND.');
  code.writeln();

  // Write part of statement.
  if (yaml.containsKey('part_of')) {
    code.writeln("part of ${yaml['part_of']};");
    code.writeln();
  }

  // Write main documentaton (description node).
  if (yaml.containsKey('description') && yaml['description'].isNotEmpty) {
    code.write(_prefixEachLine(yaml['description'], '/// '));
    code.writeln();
  }

  // Loop through messages.
  if (yaml.containsKey('messages')) {
    yaml['messages'].forEach((YamlMap msg) {
      code.write(_parseMessge(packageName, msg));
    });
  }

  // Generate formatted code.
  //
  // Note: an exception will be raised if the user specifies illegal types or
  // default values in the specification. In this case the code will fail and
  // the exception will propagate to higher-level mechanisms.
  var formatter = new DartFormatter();
  code = formatter.format(code.toString());

  // Write generated code to output file.
  return code;
}

/// Parse YAML message data.
String _parseMessge(String packageName, YamlMap msg) {
  var code = new StringBuffer();

  // Write omnibusId constant.
  var msgName = msg['name'];
  code.writeln('/// omnibusId for $msgName');
  code.writeln("const String ${msgName}Id = '$packageName.$msgName';");
  code.writeln();

  // Get message type.
  const messageClass = const {
    'event': 'OmnibusEvent',
    'request': 'OmnibusRequest',
    'response': 'OmnibusResponse'
  };
  var messageType = messageClass[(msg.containsKey('type') &&
      messageClass.containsKey(msg['type']) ? msg['type'] : 'event')];

  // Create variables for collecting data.
  var documentation = new StringBuffer();
  var positionalParameters = [];
  var optPositionalParameters = [];
  var namedParameters = [];
  var mapValues = [];

  // Add main message documentation.
  if (msg.containsKey('note') && msg['note'].isNotEmpty) {
    documentation.write(_prefixEachLine(msg['note'], '/// '));
  }

  // Can this message be initialized without checking the types?
  bool unchecked = true;

  if (msg.containsKey('properties') && msg['properties'].length > 0) {
    // Write parameter documentation heading.
    if (documentation.isNotEmpty) {
      // Note: if no message note is defined then this println should not
      // be executed.
      documentation.writeln('///');
    }
    documentation.writeln('/// *Arguments:*');

    // Loop through properties.
    List<YamlMap> properties = msg['properties'];
    properties.forEach((YamlMap param) {
      String paramName = param['name'];

      // Write parameter documentation list item.
      documentation.writeln('///'); // Blank line
      documentation.writeln('/// - **$paramName:**'); // Parameter list item

      // Write parameter documentation body.
      if (param.containsKey('note') && param['note'].isNotEmpty) {
        documentation.write(_prefixEachLine(param['note'], '///   '));
      }

      // Find out if this parameter is an enumeration.
      String type = param['type'];
      bool enumeration = type.startsWith('enum ');
      if (enumeration) {
        type = type.substring(5);
      }

      // Find out if this type required a checked message initialization.
      // Note: enumerations are stored as integer.
      // Note: if unchecked is already turned of there is no point in running
      //       the rest of this code.
      if (!enumeration && unchecked) {
        if (!_pureJsonData(type)) {
          unchecked = false;
        }
      }

      // Get default value.
      // Note: toString might not be the best way to convert the default value
      // to a Dart representation of the default value.
      String dflt = param.containsKey('dflt') ? param['dflt'].toString() : '';

      // Add parameter to correct group.
      String pass =
          param.containsKey('pass') ? param['pass'] : 'positional parameter';
      // Enforce lower-case convention.
      // Default is positional parameter.
      if (pass.contains('named')) {
        // Find standard default value is it is still empty.
        if (dflt.isEmpty) {
          if (type == 'String') {
            dflt = "''";
          } else if (type.startsWith('List<')) {
            dflt = 'const []';
          } else if (type.startsWith('Map<')) {
            dflt = 'const {}';
          } else {
            dflt = 'null';
          }
        }

        // Add parameter to named parameters.
        namedParameters.add('$type $paramName: $dflt');
      } else {
        // Add parameter to optional positional parameters if a default value is
        // defined.
        if (dflt.isEmpty) {
          positionalParameters.add('$type $paramName');
        } else {
          optPositionalParameters.add('$type $paramName = $dflt');
        }
      }

      // Add parameter to map values.
      var mapValue = enumeration ? paramName + '.index' : paramName;
      mapValues.add("'$paramName': $mapValue");
    });
  }

  // Join collected data.
  positionalParameters = positionalParameters.join(',');
  optPositionalParameters = optPositionalParameters.join(',');
  namedParameters = namedParameters.join(',');
  mapValues = mapValues.join(',');

  // Generate parameters string.
  var parameters = positionalParameters;
  if (optPositionalParameters.isNotEmpty) {
    parameters = (parameters.isNotEmpty ? parameters + ',' : '') +
        '[$optPositionalParameters]';
  }
  if (namedParameters.isNotEmpty) {
    parameters =
        (parameters.isNotEmpty ? parameters + ',' : '') + '{$namedParameters}';
  }

  // Set unchecked initialization if possible.
  var constructor = unchecked ? '$messageType.unchecked' : messageType;

  // Write message builder function. Note that the documentation is
  // newline-terminated.
  code.write('''
    $documentation$messageType $msgName($parameters) {
      return new $constructor(${msgName}Id, {$mapValues});
    }\n\n''');

  // Return generated code.
  return code.toString();
}

/// Check if the given type stringb should be validated by _getJsonValue.
///
/// Allowed types:
/// - num
/// - int
/// - double
/// - bool
/// - String
/// - List<$allowedType>
/// - Map<String, $allowedType>
bool _pureJsonData(String type) {
  var primitives = ['num', 'int', 'double', 'bool', 'String'];
  if (primitives.contains(type)) {
    return true;
  } else {
    if (type.startsWith('List<')) {
      return _pureJsonData(type.substring(5, type.length - 1));
    } else if (type.startsWith('Map<String,')) {
      return _pureJsonData(type.substring(11, type.length - 1).trimLeft());
    } else {
      return false;
    }
  }
}

/// Split [a] into lines and prefix them with [b]. Returns all lines joined
/// together using line endings.
String _prefixEachLine(String str, String prefix) {
  var out = new StringBuffer();
  var lines = LineSplitter.split(str);
  lines.forEach((String line) {
    out.writeln('$prefix$line');
  });
  return out.toString();
}
